from abc import abstractmethod

import six
import sqlalchemy as db

from dagster import check
from dagster.core.definitions.repository import RepositoryDefinition
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.scheduler import Schedule, ScheduleTick
from dagster.core.scheduler.scheduler import (
    ScheduleTickData,
    ScheduleTickStatsSnapshot,
    ScheduleTickStatus,
)
from dagster.serdes import deserialize_json_to_dagster_namedtuple, serialize_dagster_namedtuple
from dagster.utils import utc_datetime_from_timestamp

from .base import ScheduleStorage
from .schema import ScheduleTable, ScheduleTickTable


class SqlScheduleStorage(ScheduleStorage):
    '''Base class for SQL backed schedule storage
    '''

    @abstractmethod
    def connect(self):
        '''Context manager yielding a sqlalchemy.engine.Connection.'''

    def execute(self, query):
        with self.connect() as conn:
            result_proxy = conn.execute(query)
            res = result_proxy.fetchall()
            result_proxy.close()

        return res

    def _deserialize_rows(self, rows):
        return list(map(lambda r: deserialize_json_to_dagster_namedtuple(r[0]), rows))

    def all_schedules_info(self):
        '''Only for debugging purposes
        '''

        query = db.select(
            [ScheduleTable.c.repository_name, ScheduleTable.c.schedule_body]
        ).select_from(ScheduleTable)
        rows = self.execute(query)
        return list(map(lambda r: [r[0], deserialize_json_to_dagster_namedtuple(r[1])], rows))

    def all_schedules(self, repository=None):
        check.opt_inst_param(repository, 'repository', RepositoryDefinition)

        base_query = db.select([ScheduleTable.c.schedule_body]).select_from(ScheduleTable)

        if repository:
            query = base_query.where(ScheduleTable.c.repository_name == repository.name)
        else:
            query = base_query

        rows = self.execute(query)
        return self._deserialize_rows(rows)

    def get_schedule_by_name(self, repository, schedule_name):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.str_param(schedule_name, 'schedule_name')

        query = (
            db.select([ScheduleTable.c.schedule_body])
            .select_from(ScheduleTable)
            .where(ScheduleTable.c.repository_name == repository.name)
            .where(ScheduleTable.c.schedule_name == schedule_name)
        )

        rows = self.execute(query)
        return deserialize_json_to_dagster_namedtuple(rows[0][0]) if len(rows) else None

    def get_schedule_ticks_by_schedule(self, repository, schedule_name):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.str_param(schedule_name, 'schedule_name')

        query = (
            db.select([ScheduleTickTable.c.id, ScheduleTickTable.c.tick_body])
            .select_from(ScheduleTickTable)
            .where(ScheduleTickTable.c.repository_name == repository.name)
            .where(ScheduleTickTable.c.schedule_name == schedule_name)
            .order_by(ScheduleTickTable.c.id.desc())
        )

        rows = self.execute(query)
        return list(
            map(lambda r: ScheduleTick(r[0], deserialize_json_to_dagster_namedtuple(r[1])), rows)
        )

    def get_schedule_tick_stats_by_schedule(self, repository, schedule_name):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.str_param(schedule_name, 'schedule_name')

        query = (
            db.select([ScheduleTickTable.c.status, db.func.count()])
            .select_from(ScheduleTickTable)
            .where(ScheduleTickTable.c.repository_name == repository.name)
            .where(ScheduleTickTable.c.schedule_name == schedule_name)
            .group_by(ScheduleTickTable.c.status)
        )

        rows = self.execute(query)

        counts = {}
        for status, count in rows:
            counts[status] = count

        return ScheduleTickStatsSnapshot(
            ticks_started=counts.get(ScheduleTickStatus.STARTED.value, 0),
            ticks_succeeded=counts.get(ScheduleTickStatus.SUCCESS.value, 0),
            ticks_skipped=counts.get(ScheduleTickStatus.SKIPPED.value, 0),
            ticks_failed=counts.get(ScheduleTickStatus.FAILURE.value, 0),
        )

    def create_schedule_tick(self, repository, schedule_tick_data):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.inst_param(schedule_tick_data, 'schedule_tick_data', ScheduleTickData)

        with self.connect() as conn:
            try:
                tick_insert = ScheduleTickTable.insert().values(  # pylint: disable=no-value-for-parameter
                    repository_name=repository.name,
                    schedule_name=schedule_tick_data.schedule_name,
                    status=schedule_tick_data.status.value,
                    timestamp=utc_datetime_from_timestamp(schedule_tick_data.timestamp),
                    tick_body=serialize_dagster_namedtuple(schedule_tick_data),
                )
                result = conn.execute(tick_insert)
                tick_id = result.inserted_primary_key[0]
                return ScheduleTick(tick_id, schedule_tick_data)
            except db.exc.IntegrityError as exc:
                six.raise_from(
                    DagsterInvariantViolationError(
                        'Unable to insert ScheduleTick for schedule {schedule_name} in storage'.format(
                            schedule_name=schedule_tick_data.schedule_name,
                        )
                    ),
                    exc,
                )

    def update_schedule_tick(self, repository, tick):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.inst_param(tick, 'tick', ScheduleTick)

        with self.connect() as conn:
            conn.execute(
                ScheduleTickTable.update()  # pylint: disable=no-value-for-parameter
                .where(ScheduleTickTable.c.id == tick.tick_id)
                .values(
                    status=tick.status.value,
                    tick_body=serialize_dagster_namedtuple(tick.schedule_tick_data),
                )
            )

        return tick

    def add_schedule(self, repository, schedule):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.inst_param(schedule, 'schedule', Schedule)

        with self.connect() as conn:
            try:
                schedule_insert = ScheduleTable.insert().values(  # pylint: disable=no-value-for-parameter
                    repository_name=repository.name,
                    schedule_name=schedule.name,
                    status=schedule.status.value,
                    schedule_body=serialize_dagster_namedtuple(schedule),
                )
                conn.execute(schedule_insert)
            except db.exc.IntegrityError as exc:
                six.raise_from(
                    DagsterInvariantViolationError(
                        'Schedule {schedule_name} for repository {repository_name} is already present '
                        'in storage'.format(
                            schedule_name=schedule.name, repository_name=repository.name
                        )
                    ),
                    exc,
                )

        return schedule

    def update_schedule(self, repository, schedule):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.inst_param(schedule, 'schedule', Schedule)

        if not self.get_schedule_by_name(repository, schedule.name):
            raise DagsterInvariantViolationError(
                'Schedule {name} for repository {repository_name} is not present in storage'.format(
                    name=schedule.name, repository_name=repository.name
                )
            )

        with self.connect() as conn:
            conn.execute(
                ScheduleTable.update()  # pylint: disable=no-value-for-parameter
                .where(ScheduleTable.c.repository_name == repository.name)
                .where(ScheduleTable.c.schedule_name == schedule.name)
                .values(
                    status=schedule.status.value,
                    schedule_body=serialize_dagster_namedtuple(schedule),
                )
            )

    def delete_schedule(self, repository, schedule):
        check.inst_param(repository, 'repository', RepositoryDefinition)
        check.inst_param(schedule, 'schedule', Schedule)

        if not self.get_schedule_by_name(repository, schedule.name):
            raise DagsterInvariantViolationError(
                'Schedule {name} for repository {repository_name} is not present in storage'.format(
                    name=schedule.name, repository_name=repository.name
                )
            )

        with self.connect() as conn:
            conn.execute(
                ScheduleTable.delete()  # pylint: disable=no-value-for-parameter
                .where(ScheduleTable.c.repository_name == repository.name)
                .where(ScheduleTable.c.schedule_name == schedule.name)
            )

    def wipe(self):
        '''Clears the schedule storage.'''
        with self.connect() as conn:
            # https://stackoverflow.com/a/54386260/324449
            conn.execute(ScheduleTable.delete())  # pylint: disable=no-value-for-parameter
            conn.execute(ScheduleTickTable.delete())  # pylint: disable=no-value-for-parameter



from collections import defaultdict

import pytest

from dagster import (
    DagsterInvalidDefinitionError,
    DagsterTypeCheckDidNotPass,
    DependencyDefinition,
    InputDefinition,
    Int,
    List,
    Materialization,
    MultiDependencyDefinition,
    Nothing,
    Optional,
    Output,
    OutputDefinition,
    PipelineDefinition,
    SolidInvocation,
    execute_pipeline,
    lambda_solid,
    solid,
)
from dagster.core.execution.api import create_execution_plan


def _define_nothing_dep_pipeline():
    @lambda_solid(output_def=OutputDefinition(Nothing, 'complete'))
    def start_nothing():
        pass

    @lambda_solid(
        input_defs=[
            InputDefinition('add_complete', Nothing),
            InputDefinition('yield_complete', Nothing),
        ]
    )
    def end_nothing():
        pass

    @lambda_solid(output_def=OutputDefinition(Int))
    def emit_value():
        return 1

    @lambda_solid(
        input_defs=[InputDefinition('on_complete', Nothing), InputDefinition('num', Int)],
        output_def=OutputDefinition(Int),
    )
    def add_value(num):
        return 1 + num

    @solid(
        name='yield_values',
        input_defs=[InputDefinition('on_complete', Nothing)],
        output_defs=[
            OutputDefinition(Int, 'num_1'),
            OutputDefinition(Int, 'num_2'),
            OutputDefinition(Nothing, 'complete'),
        ],
    )
    def yield_values(_context):
        yield Output(1, 'num_1')
        yield Output(2, 'num_2')
        yield Output(None, 'complete')

    return PipelineDefinition(
        name='simple_exc',
        solid_defs=[emit_value, add_value, start_nothing, end_nothing, yield_values],
        dependencies={
            'add_value': {
                'on_complete': DependencyDefinition('start_nothing', 'complete'),
                'num': DependencyDefinition('emit_value'),
            },
            'yield_values': {'on_complete': DependencyDefinition('start_nothing', 'complete')},
            'end_nothing': {
                'add_complete': DependencyDefinition('add_value'),
                'yield_complete': DependencyDefinition('yield_values', 'complete'),
            },
        },
    )


def test_valid_nothing_dependencies():

    result = execute_pipeline(_define_nothing_dep_pipeline())

    assert result.success


def test_invalid_input_dependency():
    @lambda_solid(output_def=OutputDefinition(Nothing))
    def do_nothing():
        pass

    @lambda_solid(input_defs=[InputDefinition('num', Int)], output_def=OutputDefinition(Int))
    def add_one(num):
        return num + 1

    with pytest.raises(DagsterInvalidDefinitionError):
        PipelineDefinition(
            name='bad_dep',
            solid_defs=[do_nothing, add_one],
            dependencies={'add_one': {'num': DependencyDefinition('do_nothing')}},
        )


def test_result_type_check():
    @solid(output_defs=[OutputDefinition(Nothing)])
    def bad(_context):
        yield Output('oops')

    pipeline = PipelineDefinition(name='fail', solid_defs=[bad])
    with pytest.raises(DagsterTypeCheckDidNotPass):
        execute_pipeline(pipeline)


def test_nothing_inputs():
    @lambda_solid(input_defs=[InputDefinition('never_defined', Nothing)])
    def emit_one():
        return 1

    @lambda_solid
    def emit_two():
        return 2

    @lambda_solid
    def emit_three():
        return 3

    @lambda_solid(output_def=OutputDefinition(Nothing))
    def emit_nothing():
        pass

    @solid(
        input_defs=[
            InputDefinition('_one', Nothing),
            InputDefinition('one', Int),
            InputDefinition('_two', Nothing),
            InputDefinition('two', Int),
            InputDefinition('_three', Nothing),
            InputDefinition('three', Int),
        ]
    )
    def adder(_context, one, two, three):
        assert one == 1
        assert two == 2
        assert three == 3
        return one + two + three

    pipeline = PipelineDefinition(
        name='input_test',
        solid_defs=[emit_one, emit_two, emit_three, emit_nothing, adder],
        dependencies={
            SolidInvocation('emit_nothing', '_one'): {},
            SolidInvocation('emit_nothing', '_two'): {},
            SolidInvocation('emit_nothing', '_three'): {},
            'adder': {
                '_one': DependencyDefinition('_one'),
                '_two': DependencyDefinition('_two'),
                '_three': DependencyDefinition('_three'),
                'one': DependencyDefinition('emit_one'),
                'two': DependencyDefinition('emit_two'),
                'three': DependencyDefinition('emit_three'),
            },
        },
    )
    result = execute_pipeline(pipeline)
    assert result.success


def test_fanin_deps():
    called = defaultdict(int)

    @lambda_solid
    def emit_two():
        return 2

    @lambda_solid(output_def=OutputDefinition(Nothing))
    def emit_nothing():
        called['emit_nothing'] += 1

    @solid(
        input_defs=[
            InputDefinition('ready', Nothing),
            InputDefinition('num_1', Int),
            InputDefinition('num_2', Int),
        ]
    )
    def adder(_context, num_1, num_2):
        assert called['emit_nothing'] == 3
        called['adder'] += 1
        return num_1 + num_2

    pipeline = PipelineDefinition(
        name='input_test',
        solid_defs=[emit_two, emit_nothing, adder],
        dependencies={
            SolidInvocation('emit_two', 'emit_1'): {},
            SolidInvocation('emit_two', 'emit_2'): {},
            SolidInvocation('emit_nothing', '_one'): {},
            SolidInvocation('emit_nothing', '_two'): {},
            SolidInvocation('emit_nothing', '_three'): {},
            'adder': {
                'ready': MultiDependencyDefinition(
                    [
                        DependencyDefinition('_one'),
                        DependencyDefinition('_two'),
                        DependencyDefinition('_three'),
                    ]
                ),
                'num_1': DependencyDefinition('emit_1'),
                'num_2': DependencyDefinition('emit_2'),
            },
        },
    )
    result = execute_pipeline(pipeline)
    assert result.success
    assert called['adder'] == 1
    assert called['emit_nothing'] == 3


def test_valid_nothing_fns():
    @lambda_solid(output_def=OutputDefinition(Nothing))
    def just_pass():
        pass

    @solid(output_defs=[OutputDefinition(Nothing)])
    def just_pass2(_context):
        pass

    @lambda_solid(output_def=OutputDefinition(Nothing))
    def ret_none():
        return None

    @solid(output_defs=[OutputDefinition(Nothing)])
    def yield_none(_context):
        yield Output(None)

    @solid(output_defs=[OutputDefinition(Nothing)])
    def yield_stuff(_context):
        yield Materialization.file('/path/to/nowhere')

    pipeline = PipelineDefinition(
        name='fn_test', solid_defs=[just_pass, just_pass2, ret_none, yield_none, yield_stuff]
    )
    result = execute_pipeline(pipeline)
    assert result.success


def test_invalid_nothing_fns():
    @lambda_solid(output_def=OutputDefinition(Nothing))
    def ret_val():
        return 'val'

    @solid(output_defs=[OutputDefinition(Nothing)])
    def yield_val(_context):
        yield Output('val')

    with pytest.raises(DagsterTypeCheckDidNotPass):
        execute_pipeline(PipelineDefinition(name='fn_test', solid_defs=[ret_val]))

    with pytest.raises(DagsterTypeCheckDidNotPass):
        execute_pipeline(PipelineDefinition(name='fn_test', solid_defs=[yield_val]))


def test_wrapping_nothing():
    with pytest.raises(DagsterInvalidDefinitionError):

        @lambda_solid(output_def=OutputDefinition(List[Nothing]))
        def _():
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @lambda_solid(input_defs=[InputDefinition('in', List[Nothing])])
        def _(_in):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @lambda_solid(output_def=OutputDefinition(Optional[Nothing]))
        def _():
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @lambda_solid(input_defs=[InputDefinition('in', Optional[Nothing])])
        def _(_in):
            pass


def test_execution_plan():
    @solid(output_defs=[OutputDefinition(Nothing)])
    def emit_nothing(_context):
        yield Materialization.file(path='/path/')

    @lambda_solid(input_defs=[InputDefinition('ready', Nothing)])
    def consume_nothing():
        pass

    pipe = PipelineDefinition(
        name='execution_plan_test',
        solid_defs=[emit_nothing, consume_nothing],
        dependencies={'consume_nothing': {'ready': DependencyDefinition('emit_nothing')}},
    )
    plan = create_execution_plan(pipe)

    levels = plan.topological_step_levels()

    assert 'emit_nothing' in levels[0][0].key
    assert 'consume_nothing' in levels[1][0].key

    assert execute_pipeline(pipe).success



from collections import namedtuple

from dagster import check
from dagster.core.definitions.pipeline import PipelineDefinition
from dagster.core.definitions.resource import ResourceDefinition
from dagster.core.log_manager import DagsterLogManager


class InitResourceContext(
    namedtuple(
        'InitResourceContext', 'resource_config pipeline_def resource_def run_id log_manager'
    )
):
    '''Resource-specific initialization context.

    Attributes:
        resource_config (Any): The configuration data provided by the environment config. The schema
            for this data is defined by the ``config_field`` argument to
            :py:class:`ResourceDefinition`.
        pipeline_def (PipelineDefinition): The definition of the pipeline currently being executed.
        resource_def (ResourceDefinition): The definition of the resource currently being
            constructed.
        run_id (str): The id for this run of the pipeline.
        log_manager (DagsterLogManager): The log manager for this run of the pipeline
    '''

    def __new__(cls, resource_config, pipeline_def, resource_def, run_id, log_manager=None):
        return super(InitResourceContext, cls).__new__(
            cls,
            resource_config,
            check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            check.inst_param(resource_def, 'resource_def', ResourceDefinition),
            check.str_param(run_id, 'run_id'),
            check.opt_inst_param(log_manager, 'log_manager', DagsterLogManager),
        )

    @property
    def log(self):
        return self.log_manager



import re

import pytest

from dagster.check import CheckError
from dagster.seven import mock
from dagster.utils.backcompat import canonicalize_backcompat_args


def is_new(old_flag=None, new_flag=None):
    actual_new_flag = canonicalize_backcompat_args(
        new_val=new_flag,
        new_arg='new_flag',
        old_val=old_flag,
        old_arg='old_flag',
        coerce_old_to_new=lambda val: not val,
        additional_warn_txt='Will remove at next release.',
    )

    return actual_new_flag


def test_backcompat_default():
    assert is_new() is None


def test_backcompat_new_flag():
    assert is_new(new_flag=False) is False


def test_backcompat_old_flag():
    with mock.patch('warnings.warn') as warn_mock:
        assert is_new(old_flag=False) is True
        warn_mock.assert_called_once_with(
            '"old_flag" is deprecated, use "new_flag" instead. Will remove at next release.',
            stacklevel=3,
        )


def test_backcompat_both_set():
    with pytest.raises(
        CheckError,
        match=re.escape('Do not use deprecated "old_flag" now that you are using "new_flag".'),
    ):
        is_new(old_flag=False, new_flag=True)



import pytest

from dagster import (
    DependencyDefinition,
    InputDefinition,
    Int,
    OutputDefinition,
    PipelineDefinition,
    RunConfig,
    execute_pipeline,
    lambda_solid,
)
from dagster.core.errors import (
    DagsterExecutionStepNotFoundError,
    DagsterInvariantViolationError,
    DagsterRunNotFoundError,
)
from dagster.core.events import get_step_output_event
from dagster.core.execution.api import create_execution_plan, execute_plan
from dagster.core.instance import DagsterInstance
from dagster.core.storage.intermediate_store import build_fs_intermediate_store
from dagster.utils import merge_dicts


def env_with_fs(environment_dict):
    return merge_dicts(environment_dict, {'storage': {'filesystem': {}}})


def define_addy_pipeline():
    @lambda_solid(input_defs=[InputDefinition('num', Int)], output_def=OutputDefinition(Int))
    def add_one(num):
        return num + 1

    @lambda_solid(input_defs=[InputDefinition('num', Int)], output_def=OutputDefinition(Int))
    def add_two(num):
        return num + 2

    @lambda_solid(input_defs=[InputDefinition('num', Int)], output_def=OutputDefinition(Int))
    def add_three(num):
        return num + 3

    pipeline_def = PipelineDefinition(
        name='execution_plan_reexecution',
        solid_defs=[add_one, add_two, add_three],
        dependencies={
            'add_two': {'num': DependencyDefinition('add_one')},
            'add_three': {'num': DependencyDefinition('add_two')},
        },
    )
    return pipeline_def


def test_execution_plan_reexecution():
    pipeline_def = define_addy_pipeline()
    instance = DagsterInstance.ephemeral()
    environment_dict = env_with_fs({'solids': {'add_one': {'inputs': {'num': {'value': 3}}}}})
    result = execute_pipeline(pipeline_def, environment_dict=environment_dict, instance=instance,)

    assert result.success

    store = build_fs_intermediate_store(instance.intermediates_directory, result.run_id)
    assert store.get_intermediate(None, 'add_one.compute', Int).obj == 4
    assert store.get_intermediate(None, 'add_two.compute', Int).obj == 6

    ## re-execute add_two

    execution_plan = create_execution_plan(pipeline_def, environment_dict=environment_dict)

    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=pipeline_def,
        execution_plan=execution_plan,
        environment_dict=environment_dict,
        parent_run_id=result.run_id,
        root_run_id=result.run_id,
    )

    step_events = execute_plan(
        execution_plan.build_subset_plan(['add_two.compute']),
        environment_dict=environment_dict,
        pipeline_run=pipeline_run,
        instance=instance,
    )

    store = build_fs_intermediate_store(instance.intermediates_directory, pipeline_run.run_id)
    assert store.get_intermediate(None, 'add_one.compute', Int).obj == 4
    assert store.get_intermediate(None, 'add_two.compute', Int).obj == 6

    assert not get_step_output_event(step_events, 'add_one.compute')
    assert get_step_output_event(step_events, 'add_two.compute')


def test_execution_plan_wrong_run_id():
    pipeline_def = define_addy_pipeline()

    unrun_id = 'not_a_run'
    environment_dict = env_with_fs({'solids': {'add_one': {'inputs': {'num': {'value': 3}}}}})

    instance = DagsterInstance.ephemeral()

    execution_plan = create_execution_plan(pipeline_def, environment_dict=environment_dict)

    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=pipeline_def,
        execution_plan=execution_plan,
        environment_dict=environment_dict,
        parent_run_id=unrun_id,
        root_run_id=unrun_id,
    )

    with pytest.raises(DagsterRunNotFoundError) as exc_info:
        execute_plan(
            execution_plan,
            environment_dict=environment_dict,
            pipeline_run=pipeline_run,
            instance=instance,
        )

    assert str(exc_info.value) == 'Run id {} set as parent run id was not found in instance'.format(
        unrun_id
    )

    assert exc_info.value.invalid_run_id == unrun_id


def test_execution_plan_reexecution_with_in_memory():
    pipeline_def = define_addy_pipeline()
    instance = DagsterInstance.ephemeral()
    environment_dict = {'solids': {'add_one': {'inputs': {'num': {'value': 3}}}}}
    result = execute_pipeline(pipeline_def, environment_dict=environment_dict, instance=instance)

    assert result.success

    ## re-execute add_two

    execution_plan = create_execution_plan(pipeline_def, environment_dict=environment_dict)

    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=pipeline_def,
        execution_plan=execution_plan,
        environment_dict=environment_dict,
        parent_run_id=result.run_id,
        root_run_id=result.run_id,
    )

    with pytest.raises(DagsterInvariantViolationError):
        execute_plan(
            execution_plan.build_subset_plan(['add_two.compute']),
            environment_dict=environment_dict,
            pipeline_run=pipeline_run,
            instance=instance,
        )


def test_pipeline_step_key_subset_execution():
    pipeline_def = define_addy_pipeline()
    instance = DagsterInstance.ephemeral()
    environment_dict = env_with_fs({'solids': {'add_one': {'inputs': {'num': {'value': 3}}}}})
    result = execute_pipeline(pipeline_def, environment_dict=environment_dict, instance=instance,)

    assert result.success

    store = build_fs_intermediate_store(instance.intermediates_directory, result.run_id)
    assert store.get_intermediate(None, 'add_one.compute', Int).obj == 4
    assert store.get_intermediate(None, 'add_two.compute', Int).obj == 6

    ## re-execute add_two

    pipeline_reexecution_result = execute_pipeline(
        pipeline_def,
        environment_dict=environment_dict,
        run_config=RunConfig(
            previous_run_id=result.run_id, step_keys_to_execute=['add_two.compute'],
        ),
        instance=instance,
    )

    assert pipeline_reexecution_result.success

    step_events = pipeline_reexecution_result.step_event_list
    assert step_events

    store = build_fs_intermediate_store(
        instance.intermediates_directory, pipeline_reexecution_result.run_id
    )
    assert store.get_intermediate(None, 'add_one.compute', Int).obj == 4
    assert store.get_intermediate(None, 'add_two.compute', Int).obj == 6

    assert not get_step_output_event(step_events, 'add_one.compute')
    assert get_step_output_event(step_events, 'add_two.compute')


def test_pipeline_step_key_subset_execution_wrong_step_key_in_subset():
    pipeline_def = define_addy_pipeline()
    environment_dict = env_with_fs({'solids': {'add_one': {'inputs': {'num': {'value': 3}}}}})
    result = execute_pipeline(pipeline_def, environment_dict=environment_dict)
    assert result.success

    with pytest.raises(DagsterExecutionStepNotFoundError):
        execute_pipeline(
            pipeline_def,
            environment_dict=environment_dict,
            run_config=RunConfig(previous_run_id=result.run_id, step_keys_to_execute=['nope']),
        )



import os
import sys

import click

from dagster.cli.load_handle import handle_for_repo_cli_args
from dagster.cli.pipeline import repository_target_argument
from dagster.core.snap import active_repository_data_from_def
from dagster.serdes import serialize_dagster_namedtuple


def create_repository_cli_group():
    group = click.Group(name='repository')
    group.add_command(snapshot_command)
    return group


@click.command(
    name='snapshot',
    help='Snapshot the given repository definition and load into the serialization target.',
)
@click.argument('output_file', type=click.Path())
@repository_target_argument
def snapshot_command(output_file, **kwargs):
    handle = handle_for_repo_cli_args(kwargs)

    # add the path for the cwd so imports in dynamically loaded code work correctly
    sys.path.append(os.getcwd())

    definition = handle.entrypoint.perform_load()
    active_repo_data = active_repository_data_from_def(definition)
    with open(os.path.abspath(output_file), 'w+') as fp:
        fp.write(serialize_dagster_namedtuple(active_repo_data))


repository_cli = create_repository_cli_group()



from dagster.builtins import Any, Bool, Float, Int, Nothing, String
from dagster.config import Enum, EnumValue, Field, Permissive, Selector, Shape
from dagster.config.config_type import Array, Noneable, ScalarUnion
from dagster.core.definitions import (
    CompositeSolidDefinition,
    ConfigMapping,
    DependencyDefinition,
    EventMetadataEntry,
    ExecutionTargetHandle,
    ExecutorDefinition,
    ExpectationResult,
    Failure,
    InputDefinition,
    InputMapping,
    JsonMetadataEntryData,
    LoggerDefinition,
    MarkdownMetadataEntryData,
    Materialization,
    ModeDefinition,
    MultiDependencyDefinition,
    Output,
    OutputDefinition,
    OutputMapping,
    Partition,
    PartitionSetDefinition,
    PathMetadataEntryData,
    PipelineDefinition,
    PresetDefinition,
    RepositoryDefinition,
    ResourceDefinition,
    RetryRequested,
    ScheduleDefinition,
    ScheduleExecutionContext,
    SolidDefinition,
    SolidInvocation,
    SystemStorageData,
    SystemStorageDefinition,
    TextMetadataEntryData,
    TypeCheck,
    UrlMetadataEntryData,
    composite_solid,
    daily_schedule,
    default_executors,
    executor,
    hourly_schedule,
    in_process_executor,
    lambda_solid,
    logger,
    monthly_schedule,
    multiprocess_executor,
    pipeline,
    repository_partitions,
    resource,
    schedule,
    schedules,
    solid,
    system_storage,
    weekly_schedule,
)
from dagster.core.engine import Engine
from dagster.core.engine.init import InitExecutorContext
from dagster.core.errors import (
    DagsterConfigMappingFunctionError,
    DagsterError,
    DagsterEventLogInvalidForRun,
    DagsterExecutionStepExecutionError,
    DagsterExecutionStepNotFoundError,
    DagsterInvalidConfigDefinitionError,
    DagsterInvalidConfigError,
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    DagsterResourceFunctionError,
    DagsterRunNotFoundError,
    DagsterStepOutputNotFoundError,
    DagsterSubprocessError,
    DagsterTypeCheckDidNotPass,
    DagsterTypeCheckError,
    DagsterUnknownResourceError,
    DagsterUnmetExecutorRequirementsError,
    DagsterUserCodeExecutionError,
)
from dagster.core.events import DagsterEvent, DagsterEventType
from dagster.core.execution.api import (
    execute_partition_set,
    execute_pipeline,
    execute_pipeline_iterator,
)
from dagster.core.execution.config import ExecutorConfig, RunConfig
from dagster.core.execution.context.compute import SolidExecutionContext
from dagster.core.execution.context.init import InitResourceContext
from dagster.core.execution.context.logger import InitLoggerContext
from dagster.core.execution.context.system import SystemComputeExecutionContext, TypeCheckContext
from dagster.core.execution.results import (
    CompositeSolidExecutionResult,
    PipelineExecutionResult,
    SolidExecutionResult,
)
from dagster.core.instance import DagsterInstance
from dagster.core.log_manager import DagsterLogManager
from dagster.core.storage.file_manager import FileHandle, LocalFileHandle
from dagster.core.storage.init import InitSystemStorageContext
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.core.storage.system_storage import (
    default_system_storage_defs,
    fs_system_storage,
    mem_system_storage,
)
from dagster.core.types.config_schema import input_hydration_config, output_materialization_config
from dagster.core.types.dagster_type import (
    DagsterType,
    List,
    Optional,
    Path,
    PythonObjectDagsterType,
)
from dagster.core.types.decorator import (
    make_python_type_usable_as_dagster_type,
    usable_as_dagster_type,
)
from dagster.core.types.marshal import SerializationStrategy
from dagster.core.types.python_dict import Dict
from dagster.core.types.python_set import Set
from dagster.core.types.python_tuple import Tuple
from dagster.utils import file_relative_path
from dagster.utils.test import (
    check_dagster_type,
    execute_solid,
    execute_solid_within_pipeline,
    execute_solids_within_pipeline,
)

from .version import __version__

from dagster.config.source import StringSource, IntSource  # isort:skip


__all__ = [
    # Definition
    'CompositeSolidDefinition',
    'ConfigMapping',
    'DependencyDefinition',
    'EventMetadataEntry',
    'ExecutorDefinition',
    'ExpectationResult',
    'Failure',
    'Field',
    'InputDefinition',
    'InputMapping',
    'JsonMetadataEntryData',
    'LoggerDefinition',
    'MarkdownMetadataEntryData',
    'Materialization',
    'ModeDefinition',
    'MultiDependencyDefinition',
    'Output',
    'OutputDefinition',
    'OutputMapping',
    'PathMetadataEntryData',
    'PipelineDefinition',
    'PresetDefinition',
    'RepositoryDefinition',
    'ResourceDefinition',
    'SolidDefinition',
    'SolidInvocation',
    'SystemStorageDefinition',
    'TextMetadataEntryData',
    'UrlMetadataEntryData',
    # Decorators
    'composite_solid',
    'executor',
    'lambda_solid',
    'logger',
    'pipeline',
    'resource',
    'schedule',
    'schedules',
    'solid',
    'system_storage',
    # Execution
    'CompositeSolidExecutionResult',
    'DagsterEvent',
    'DagsterEventType',
    'Engine',
    'ExecutionTargetHandle',
    'ExecutorConfig',
    'InitExecutorContext',
    'InitLoggerContext',
    'InitResourceContext',
    'InitSystemStorageContext',
    'PipelineExecutionResult',
    'RetryRequested',
    'RunConfig',
    'SolidExecutionResult',
    'SystemComputeExecutionContext',
    'SolidExecutionContext',
    'SystemStorageData',
    'TypeCheckContext',
    'PipelineRun',
    'default_executors',
    'default_system_storage_defs',
    'execute_pipeline_iterator',
    'execute_pipeline',
    'execute_solid_within_pipeline',
    'fs_system_storage',
    'in_process_executor',
    'mem_system_storage',
    'multiprocess_executor',
    # Errors
    'DagsterConfigMappingFunctionError',
    'DagsterError',
    'DagsterEventLogInvalidForRun',
    'DagsterExecutionStepExecutionError',
    'DagsterExecutionStepNotFoundError',
    'DagsterInvalidConfigDefinitionError',
    'DagsterInvalidConfigError',
    'DagsterInvalidDefinitionError',
    'DagsterInvariantViolationError',
    'DagsterResourceFunctionError',
    'DagsterRunNotFoundError',
    'DagsterStepOutputNotFoundError',
    'DagsterSubprocessError',
    'DagsterTypeCheckDidNotPass',
    'DagsterTypeCheckError',
    'DagsterUnknownResourceError',
    'DagsterUnmetExecutorRequirementsError',
    'DagsterUserCodeExecutionError',
    # Logging
    'DagsterLogManager',
    # Utilities
    'check_dagster_type',
    'execute_solid',
    'execute_solids_within_pipeline',
    'file_relative_path',
    # types
    'Any',
    'Bool',
    'Dict',
    'Enum',
    'EnumValue',
    'Float',
    'Int',
    'List',
    'Nothing',
    'Optional',
    'Path',
    'SerializationStrategy',
    'Set',
    'String',
    'Tuple',
    'TypeCheck',
    'input_hydration_config',
    'output_materialization_config',
    # type creation
    'DagsterType',
    'PythonObjectDagsterType',
    'make_python_type_usable_as_dagster_type',
    'usable_as_dagster_type',
    # config
    'Array',
    'Noneable',
    'Permissive',
    'ScalarUnion',
    'StringSource',
    'IntSource',
    'Selector',
    'Shape',
    # file things
    'FileHandle',
    'LocalFileHandle',
    # instance
    'DagsterInstance',
    # partitions and schedules
    'Partition',
    'PartitionSetDefinition',
    'ScheduleDefinition',
    'ScheduleExecutionContext',
    'daily_schedule',
    'execute_partition_set',
    'hourly_schedule',
    'monthly_schedule',
    'weekly_schedule',
    'repository_partitions',
]



import os
import pickle
import sys

from dagster.core.execution.plan.external_step import PICKLED_EVENTS_FILE_NAME, run_step_from_ref
from dagster.core.storage.file_manager import LocalFileHandle, LocalFileManager


def main(step_run_ref_path):
    file_manager = LocalFileManager('.')
    file_handle = LocalFileHandle(step_run_ref_path)
    step_run_ref = pickle.loads(file_manager.read_data(file_handle))

    events = list(run_step_from_ref(step_run_ref))
    events_out_path = os.path.join(os.path.dirname(step_run_ref_path), PICKLED_EVENTS_FILE_NAME)
    with open(events_out_path, 'wb') as events_file:
        pickle.dump(events, events_file)


if __name__ == '__main__':
    main(sys.argv[1])



from dagster import Int, logger


def test_dagster_type_logger_decorator_config():
    @logger(Int)
    def dagster_type_logger_config(_):
        raise Exception('not called')

    assert dagster_type_logger_config.config_field.config_type.given_name == 'Int'

    @logger(int)
    def python_type_logger_config(_):
        raise Exception('not called')

    assert python_type_logger_config.config_field.config_type.given_name == 'Int'



from enum import Enum

from dagster.utils import is_enum_value


class APythonEnum(Enum):
    VAL = 'VAL'


def test_enum_value():
    assert is_enum_value(None) is False
    assert is_enum_value(1) is False
    assert is_enum_value('foo') is False
    assert is_enum_value('VAL') is False
    assert is_enum_value(APythonEnum.VAL) is True



from .scheduler import (
    Schedule,
    ScheduleDefinitionData,
    ScheduleStatus,
    ScheduleTick,
    ScheduleTickStatus,
    Scheduler,
    SchedulerHandle,
    get_schedule_change_set,
    reconcile_scheduler_state,
)



import os

import click

from dagster import check
from dagster.core.instance import DagsterInstance


def create_instance_cli_group():
    group = click.Group(name='instance')
    group.add_command(info_command)
    group.add_command(migrate_command)
    return group


@click.command(name='info', help='List the information about the current instance.')
def info_command():
    instance = DagsterInstance.get()
    home = os.environ.get('DAGSTER_HOME')

    if instance.is_ephemeral:
        check.invariant(
            home is None,
            'Unexpected state, ephemeral instance but DAGSTER_HOME is set to "{}"'.format(home),
        )
        click.echo(
            '$DAGSTER_HOME is not set, using an ephemeral instance. '
            'Run artifacts will only exist in memory and any filesystem access '
            'will use a temp directory that gets cleaned up on exit.'
        )
        return

    click.echo('$DAGSTER_HOME: {}\n'.format(home))

    click.echo(instance.info_str())


@click.command(name='migrate', help='Automatically migrate an out of date instance.')
def migrate_command():
    instance = DagsterInstance.get()
    home = os.environ.get('DAGSTER_HOME')

    if instance.is_ephemeral:
        click.echo('$DAGSTER_HOME is not set; ephemeral instances do not need to be migrated.')
        return

    click.echo('$DAGSTER_HOME: {}\n'.format(home))

    instance.upgrade(click.echo)

    click.echo(instance.info_str())


instance_cli = create_instance_cli_group()



"""add step_key pipeline_name

Revision ID: 3b1e175a2be3
Revises: 567bc23fd1ac
Create Date: 2020-03-31 11:01:42.609069

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import reflection

# pylint: disable=no-member

# revision identifiers, used by Alembic.
revision = '3b1e175a2be3'
down_revision = '567bc23fd1ac'
branch_labels = None
depends_on = None


def upgrade():
    bind = op.get_context().bind
    inspector = reflection.Inspector.from_engine(bind)
    has_tables = inspector.get_table_names()
    if 'event_logs' in has_tables:
        op.add_column('event_logs', sa.Column('step_key', sa.String))


def downgrade():
    bind = op.get_context().bind
    inspector = reflection.Inspector.from_engine(bind)
    has_tables = inspector.get_table_names()
    if 'event_logs' in has_tables:
        op.drop_column('event_logs', 'step_key')



import os
import time
from threading import Thread

import pytest

from dagster import (
    DagsterEventType,
    DagsterSubprocessError,
    ExecutionTargetHandle,
    Field,
    ModeDefinition,
    String,
    execute_pipeline_iterator,
    pipeline,
    resource,
    seven,
    solid,
)
from dagster.core.instance import DagsterInstance
from dagster.utils import safe_tempfile_path


def _send_kbd_int(temp_files):
    while not all([os.path.exists(temp_file) for temp_file in temp_files]):
        time.sleep(0.1)
    seven.thread.interrupt_main()


@solid(config={'tempfile': Field(String)})
def write_a_file(context):
    with open(context.solid_config['tempfile'], 'w') as ff:
        ff.write('yup')

    while True:
        time.sleep(0.1)


@solid
def should_not_start(_context):
    assert False


@pipeline
def write_files_pipeline():
    write_a_file.alias('write_1')()
    write_a_file.alias('write_2')()
    write_a_file.alias('write_3')()
    write_a_file.alias('write_4')()
    should_not_start.alias('x_should_not_start')()
    should_not_start.alias('y_should_not_start')()
    should_not_start.alias('z_should_not_start')()


def test_interrupt():
    @pipeline
    def write_a_file_pipeline():
        write_a_file()

    with safe_tempfile_path() as success_tempfile:

        # launch a thread the waits until the file is written to launch an interrupt
        Thread(target=_send_kbd_int, args=([success_tempfile],)).start()

        results = []
        try:
            # launch a pipeline that writes a file and loops infinitely
            # next time the launched thread wakes up it will send a keyboard
            # interrupt
            for result in execute_pipeline_iterator(
                write_a_file_pipeline,
                environment_dict={
                    'solids': {'write_a_file': {'config': {'tempfile': success_tempfile}}}
                },
            ):
                results.append(result.event_type)
            assert False  # should never reach
        except KeyboardInterrupt:
            pass

        assert DagsterEventType.STEP_FAILURE in results
        assert DagsterEventType.PIPELINE_FAILURE in results


@pytest.mark.skip('https://github.com/dagster-io/dagster/issues/1970')
def test_interrupt_multiproc():
    with seven.TemporaryDirectory() as tempdir:
        file_1 = os.path.join(tempdir, 'file_1')
        file_2 = os.path.join(tempdir, 'file_2')
        file_3 = os.path.join(tempdir, 'file_3')
        file_4 = os.path.join(tempdir, 'file_4')

        # launch a thread the waits until the file is written to launch an interrupt
        Thread(target=_send_kbd_int, args=([file_1, file_2, file_3, file_4],)).start()

        results = []
        try:
            # launch a pipeline that writes a file and loops infinitely
            # next time the launched thread wakes up it will send a keyboard
            # interrupt
            for result in execute_pipeline_iterator(
                ExecutionTargetHandle.for_pipeline_python_file(
                    __file__, 'write_files_pipeline'
                ).build_pipeline_definition(),
                environment_dict={
                    'solids': {
                        'write_1': {'config': {'tempfile': file_1}},
                        'write_2': {'config': {'tempfile': file_2}},
                        'write_3': {'config': {'tempfile': file_3}},
                        'write_4': {'config': {'tempfile': file_4}},
                    },
                    'execution': {'multiprocess': {'config': {'max_concurrent': 4}}},
                    'storage': {'filesystem': {}},
                },
                instance=DagsterInstance.local_temp(tempdir=tempdir),
            ):
                results.append(result)
            assert False  # should never reach
        except (DagsterSubprocessError, KeyboardInterrupt):
            pass

        assert [result.event_type for result in results].count(DagsterEventType.STEP_FAILURE) == 4
        assert DagsterEventType.PIPELINE_FAILURE in [result.event_type for result in results]


def test_interrupt_resource_teardown():
    called = []
    cleaned = []

    @resource
    def resource_a(_):
        try:
            called.append('A')
            yield 'A'
        finally:
            cleaned.append('A')

    @solid(config={'tempfile': Field(String)}, required_resource_keys={'a'})
    def write_a_file_resource_solid(context):
        with open(context.solid_config['tempfile'], 'w') as ff:
            ff.write('yup')

        while True:
            time.sleep(0.1)

    @pipeline(mode_defs=[ModeDefinition(resource_defs={'a': resource_a})])
    def write_a_file_pipeline():
        write_a_file_resource_solid()

    with safe_tempfile_path() as success_tempfile:

        # launch a thread the waits until the file is written to launch an interrupt
        Thread(target=_send_kbd_int, args=([success_tempfile],)).start()

        results = []
        try:
            # launch a pipeline that writes a file and loops infinitely
            # next time the launched thread wakes up it will send a keyboard
            # interrupt
            for result in execute_pipeline_iterator(
                write_a_file_pipeline,
                environment_dict={
                    'solids': {
                        'write_a_file_resource_solid': {'config': {'tempfile': success_tempfile}}
                    }
                },
            ):
                results.append(result.event_type)
            assert False  # should never reach
        except KeyboardInterrupt:
            pass

        assert DagsterEventType.STEP_FAILURE in results
        assert DagsterEventType.PIPELINE_FAILURE in results
        assert 'A' in cleaned



from dagster import Field, Int, String, composite_solid, pipeline, solid
from dagster.core.execution.api import create_execution_plan, execute_plan
from dagster.core.instance import DagsterInstance


@solid(config={'foo': Field(String)})
def node_a(context):
    return context.solid_config['foo']


@solid(config={'bar': Int})
def node_b(context, input_):
    return input_ * context.solid_config['bar']


@composite_solid
def composite_with_nested_config_solid():
    return node_b(node_a())


@pipeline
def composite_pipeline():
    return composite_with_nested_config_solid()


@composite_solid(
    config_fn=lambda cfg: {
        'node_a': {'config': {'foo': cfg['foo']}},
        'node_b': {'config': {'bar': cfg['bar']}},
    },
    config={'foo': Field(String), 'bar': Int},
)
def composite_with_nested_config_solid_and_config_mapping():
    return node_b(node_a())


@pipeline
def composite_pipeline_with_config_mapping():
    return composite_with_nested_config_solid_and_config_mapping()


def test_execution_plan_for_composite_solid():
    environment_dict = {
        'solids': {
            'composite_with_nested_config_solid': {
                'solids': {'node_a': {'config': {'foo': 'baz'}}, 'node_b': {'config': {'bar': 3}}}
            }
        }
    }
    execution_plan = create_execution_plan(composite_pipeline, environment_dict=environment_dict)
    instance = DagsterInstance.ephemeral()
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=composite_pipeline, execution_plan=execution_plan
    )
    events = execute_plan(
        execution_plan,
        environment_dict=environment_dict,
        pipeline_run=pipeline_run,
        instance=instance,
    )

    assert [e.event_type_value for e in events] == [
        'STEP_START',
        'STEP_OUTPUT',
        'STEP_SUCCESS',
        'STEP_START',
        'STEP_INPUT',
        'STEP_OUTPUT',
        'STEP_SUCCESS',
    ]


def test_execution_plan_for_composite_solid_with_config_mapping():
    environment_dict = {
        'solids': {
            'composite_with_nested_config_solid_and_config_mapping': {
                'config': {'foo': 'baz', 'bar': 3}
            }
        }
    }
    execution_plan = create_execution_plan(
        composite_pipeline_with_config_mapping, environment_dict=environment_dict
    )
    instance = DagsterInstance.ephemeral()
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=composite_pipeline_with_config_mapping, execution_plan=execution_plan
    )

    events = execute_plan(
        execution_plan,
        environment_dict=environment_dict,
        pipeline_run=pipeline_run,
        instance=instance,
    )

    assert [e.event_type_value for e in events] == [
        'STEP_START',
        'STEP_OUTPUT',
        'STEP_SUCCESS',
        'STEP_START',
        'STEP_INPUT',
        'STEP_OUTPUT',
        'STEP_SUCCESS',
    ]



import pytest
from click.testing import CliRunner

from dagster import file_relative_path
from dagster.check import ParameterCheckError
from dagster.cli.api import pipeline_snapshot_command, repository_snapshot_command
from dagster.core.snap import ActiveRepositoryData
from dagster.core.snap.active_data import ActivePipelineData
from dagster.serdes import deserialize_json_to_dagster_namedtuple


def test_snapshot_command_repository():
    runner = CliRunner()
    result = runner.invoke(
        repository_snapshot_command, ['-y', file_relative_path(__file__, 'repository_file.yaml')],
    )
    assert result.exit_code == 0
    # Now that we have the snapshot make sure that it can be properly deserialized
    active_repository_data = deserialize_json_to_dagster_namedtuple(result.output)
    assert isinstance(active_repository_data, ActiveRepositoryData)
    assert active_repository_data.name == 'bar'
    assert len(active_repository_data.active_pipeline_datas) == 2


def test_snapshot_command_repository_fail_on_pipeline_handle():
    runner = CliRunner()
    with pytest.raises(ParameterCheckError):
        result = runner.invoke(
            repository_snapshot_command,
            ['-f', file_relative_path(__file__, 'test_cli_commands.py'), '-n', 'baz_pipeline',],
        )
        assert result.exit_code == 1
        raise result.exception


def test_snapshot_command_pipeline():
    runner = CliRunner()
    result = runner.invoke(
        pipeline_snapshot_command,
        ['-y', file_relative_path(__file__, 'repository_file.yaml'), 'foo'],
    )
    assert result.exit_code == 0
    # Now that we have the snapshot make sure that it can be properly deserialized
    active_pipeline_data = deserialize_json_to_dagster_namedtuple(result.output)
    assert isinstance(active_pipeline_data, ActivePipelineData)
    assert active_pipeline_data.name == 'foo'
    assert (
        len(active_pipeline_data.pipeline_snapshot.solid_definitions_snapshot.solid_def_snaps) == 2
    )


def test_snapshot_command_pipeline_solid_subset():
    runner = CliRunner()
    result = runner.invoke(
        pipeline_snapshot_command,
        [
            '-y',
            file_relative_path(__file__, 'repository_file.yaml'),
            'foo',
            '--solid-subset',
            'do_input',
        ],
    )
    assert result.exit_code == 0
    # Now that we have the snapshot make sure that it can be properly deserialized
    active_pipeline_data = deserialize_json_to_dagster_namedtuple(result.output)
    assert isinstance(active_pipeline_data, ActivePipelineData)
    assert active_pipeline_data.name == 'foo'
    assert (
        len(active_pipeline_data.pipeline_snapshot.solid_definitions_snapshot.solid_def_snaps) == 1
    )



import pytest

from dagster import String
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.core.types.config_schema import input_hydration_config


def test_input_hydration_config_one():
    @input_hydration_config(String)
    def _foo(_, hello):
        return hello


def test_input_hydration_config_missing_context():

    with pytest.raises(DagsterInvalidDefinitionError):

        @input_hydration_config(String)
        def _foo(hello):
            return hello


def test_input_hydration_config_missing_variable():

    with pytest.raises(DagsterInvalidDefinitionError):

        @input_hydration_config(String)
        def _foo(_):
            return 1



import importlib

import pytest

from dagster.cli.load_handle import UsageError, handle_for_pipeline_cli_args
from dagster.core.definitions import (
    ExecutionTargetHandle,
    LoaderEntrypoint,
    PipelineDefinition,
    RepositoryDefinition,
)
from dagster.core.definitions.decorators import lambda_solid
from dagster.core.definitions.handle import _ExecutionTargetMode
from dagster.seven import import_module_from_path
from dagster.utils import file_relative_path


def test_repository_python_file():
    python_file = file_relative_path(__file__, 'bar_repo.py')
    module = import_module_from_path('bar_repo', python_file)

    handle = handle_for_pipeline_cli_args(
        {'pipeline_name': 'foo', 'python_file': python_file, 'fn_name': 'define_bar_repo'}
    )
    assert handle.mode == _ExecutionTargetMode.REPOSITORY
    assert handle.entrypoint == LoaderEntrypoint(module, 'bar_repo', 'define_bar_repo', handle)
    assert handle.data.pipeline_name == 'foo'
    assert handle.entrypoint.from_handle == handle

    with pytest.raises(UsageError):
        handle_for_pipeline_cli_args(
            {
                'module_name': 'kdjfkd',
                'pipeline_name': 'foo',
                'python_file': file_relative_path(__file__, 'bar_repo.py'),
                'fn_name': 'define_bar_repo',
                'repository_yaml': None,
            }
        )

    with pytest.raises(UsageError):
        handle_for_pipeline_cli_args(
            {
                'module_name': None,
                'pipeline_name': 'foo',
                'python_file': file_relative_path(__file__, 'bar_repo.py'),
                'fn_name': 'define_bar_repo',
                'repository_yaml': 'kjdfkdjf',
            }
        )


def test_repository_module():
    module = importlib.import_module('dagster')

    handle = handle_for_pipeline_cli_args(
        {
            'module_name': 'dagster',
            'pipeline_name': 'foo',
            'python_file': None,
            'fn_name': 'define_bar_repo',
            'repository_yaml': None,
        }
    )
    assert handle.mode == _ExecutionTargetMode.REPOSITORY
    expected = LoaderEntrypoint(module, 'dagster', 'define_bar_repo')
    assert handle.entrypoint.module == expected.module
    assert handle.entrypoint.module_name == expected.module_name
    assert handle.entrypoint.fn_name == expected.fn_name
    assert handle.entrypoint.from_handle == handle
    assert handle.data.pipeline_name == 'foo'


def test_pipeline_python_file():
    python_file = file_relative_path(__file__, 'foo_pipeline.py')
    module = import_module_from_path('foo_pipeline', python_file)

    handle = handle_for_pipeline_cli_args(
        {
            'module_name': None,
            'fn_name': 'define_pipeline',
            'pipeline_name': None,
            'python_file': python_file,
            'repository_yaml': None,
        }
    )
    assert handle.mode == _ExecutionTargetMode.PIPELINE
    expected = LoaderEntrypoint(module, 'foo_pipeline', 'define_pipeline')
    assert handle.entrypoint.module == expected.module
    assert handle.entrypoint.module_name == expected.module_name
    assert handle.entrypoint.fn_name == expected.fn_name
    assert handle.entrypoint.from_handle == handle


def test_pipeline_module():
    handle = handle_for_pipeline_cli_args(
        {
            'module_name': 'dagster',
            'fn_name': 'define_pipeline',
            'pipeline_name': None,
            'python_file': None,
            'repository_yaml': None,
        }
    )
    assert handle.mode == _ExecutionTargetMode.PIPELINE
    expected = LoaderEntrypoint(importlib.import_module('dagster'), 'dagster', 'define_pipeline')
    assert handle.entrypoint.module == expected.module
    assert handle.entrypoint.module_name == expected.module_name
    assert handle.entrypoint.fn_name == expected.fn_name
    assert handle.entrypoint.from_handle == handle


def test_yaml_file():
    module = importlib.import_module('dagster_examples.intro_tutorial.repos')

    handle = handle_for_pipeline_cli_args(
        {
            'module_name': None,
            'pipeline_name': 'foobar',
            'python_file': None,
            'fn_name': None,
            'repository_yaml': file_relative_path(__file__, 'repository_module.yaml'),
        }
    )
    assert handle.mode == _ExecutionTargetMode.REPOSITORY

    expected = LoaderEntrypoint(module, 'dagster_examples.intro_tutorial.repos', 'define_repo')
    assert handle.entrypoint.module == expected.module
    assert handle.entrypoint.module_name == expected.module_name

    assert handle.entrypoint.fn_name == expected.fn_name
    assert handle.entrypoint.from_handle == handle

    assert handle.data.pipeline_name == 'foobar'

    with pytest.raises(UsageError):
        assert handle_for_pipeline_cli_args({'module_name': 'kdjfdk', 'pipeline_name': 'foobar'})

    with pytest.raises(UsageError):
        assert handle_for_pipeline_cli_args({'fn_name': 'kjdfkd', 'pipeline_name': 'foobar'})

    with pytest.raises(UsageError):
        assert handle_for_pipeline_cli_args({'pipeline_name': 'foobar', 'python_file': 'kjdfkdj'})


def test_load_from_repository_file():
    handle = handle_for_pipeline_cli_args(
        {'pipeline_name': 'foo', 'python_file': __file__, 'fn_name': 'define_bar_repo'}
    )
    pipeline = handle.build_pipeline_definition()

    assert isinstance(pipeline, PipelineDefinition)
    assert pipeline.name == 'foo'

    assert ExecutionTargetHandle.get_handle(pipeline) == (handle, None)


def test_load_from_repository_module():
    handle = handle_for_pipeline_cli_args(
        {
            'module_name': 'dagster_examples.intro_tutorial.repos',
            'pipeline_name': 'hello_cereal_pipeline',
            'fn_name': 'define_repo',
        }
    )
    pipeline = handle.build_pipeline_definition()

    assert isinstance(pipeline, PipelineDefinition)
    assert pipeline.name == 'hello_cereal_pipeline'
    assert ExecutionTargetHandle.get_handle(pipeline) == (handle, None)


def test_load_from_pipeline_file():
    handle = handle_for_pipeline_cli_args(
        {'fn_name': 'define_foo_pipeline', 'python_file': __file__}
    )
    pipeline = handle.build_pipeline_definition()

    assert isinstance(pipeline, PipelineDefinition)
    assert pipeline.name == 'foo'
    assert ExecutionTargetHandle.get_handle(pipeline) == (handle, None)


def test_load_from_pipeline_module():
    handle = handle_for_pipeline_cli_args(
        {'module_name': 'dagster_examples.intro_tutorial.repos', 'fn_name': 'hello_cereal_pipeline'}
    )
    pipeline = handle.build_pipeline_definition()

    assert isinstance(pipeline, PipelineDefinition)
    assert pipeline.name == 'hello_cereal_pipeline'
    assert ExecutionTargetHandle.get_handle(pipeline) == (handle, None)


def test_loader_from_default_repository_module_yaml():
    handle = handle_for_pipeline_cli_args(
        {
            'pipeline_name': 'hello_cereal_pipeline',
            'repository_yaml': file_relative_path(__file__, 'repository_module.yaml'),
        }
    )
    pipeline = handle.build_pipeline_definition()

    assert isinstance(pipeline, PipelineDefinition)
    assert pipeline.name == 'hello_cereal_pipeline'
    assert ExecutionTargetHandle.get_handle(pipeline) == (handle, None)


def test_loader_from_default_repository_file_yaml():
    handle = handle_for_pipeline_cli_args(
        {
            'pipeline_name': 'foo',
            'repository_yaml': file_relative_path(__file__, 'repository_file.yaml'),
        }
    )
    pipeline = handle.build_pipeline_definition()

    assert isinstance(pipeline, PipelineDefinition)
    assert pipeline.name == 'foo'
    assert ExecutionTargetHandle.get_handle(pipeline) == (handle, None)


def define_foo_pipeline():
    @lambda_solid
    def do_something():
        return 1

    return PipelineDefinition(name='foo', solid_defs=[do_something])


def define_bar_repo():
    return RepositoryDefinition('bar', {'foo': define_foo_pipeline})



import os
import time

from dagster import (
    ExecutionTargetHandle,
    Field,
    InputDefinition,
    Nothing,
    OutputDefinition,
    PresetDefinition,
    String,
    execute_pipeline,
    lambda_solid,
    pipeline,
    solid,
)
from dagster.core.instance import DagsterInstance
from dagster.utils import safe_tempfile_path


def test_diamond_simple_execution():
    result = execute_pipeline(define_diamond_pipeline())
    assert result.success
    assert result.result_for_solid('adder').output_value() == 11


def compute_event(result, solid_name):
    return result.result_for_solid(solid_name).compute_step_events[0]


def test_diamond_multi_execution():
    pipe = ExecutionTargetHandle.for_pipeline_python_file(
        __file__, 'define_diamond_pipeline'
    ).build_pipeline_definition()
    result = execute_pipeline(
        pipe,
        environment_dict={'storage': {'filesystem': {}}, 'execution': {'multiprocess': {}}},
        instance=DagsterInstance.local_temp(),
    )
    assert result.success

    assert result.result_for_solid('adder').output_value() == 11

    # https://github.com/dagster-io/dagster/issues/1875
    # pids_by_solid = {}
    # for solid in pipeline.solids:
    #     pids_by_solid[solid.name] = compute_event(result, solid.name).logging_tags['pid']

    # # guarantee that all solids ran in their own process
    # assert len(set(pids_by_solid.values())) == len(pipeline.solids)


def define_diamond_pipeline():
    @lambda_solid
    def return_two():
        return 2

    @lambda_solid(input_defs=[InputDefinition('num')])
    def add_three(num):
        return num + 3

    @lambda_solid(input_defs=[InputDefinition('num')])
    def mult_three(num):
        return num * 3

    @lambda_solid(input_defs=[InputDefinition('left'), InputDefinition('right')])
    def adder(left, right):
        return left + right

    @pipeline(
        preset_defs=[
            PresetDefinition(
                'just_adder',
                {
                    'storage': {'filesystem': {}},
                    'execution': {'multiprocess': {}},
                    'solids': {'adder': {'inputs': {'left': {'value': 1}, 'right': {'value': 1}}}},
                },
                solid_subset=['adder'],
            )
        ],
    )
    def diamond_pipeline():
        two = return_two()
        adder(left=add_three(two), right=mult_three(two))

    return diamond_pipeline


def define_error_pipeline():
    @lambda_solid
    def should_never_execute(_x):
        assert False  # this should never execute

    @lambda_solid
    def throw_error():
        raise Exception('bad programmer')

    @pipeline
    def error_pipeline():
        should_never_execute(throw_error())

    return error_pipeline


def test_error_pipeline():
    pipe = define_error_pipeline()
    result = execute_pipeline(pipe, raise_on_error=False)
    assert not result.success


def test_error_pipeline_multiprocess():
    result = execute_pipeline(
        ExecutionTargetHandle.for_pipeline_fn(define_error_pipeline).build_pipeline_definition(),
        environment_dict={'storage': {'filesystem': {}}, 'execution': {'multiprocess': {}}},
        instance=DagsterInstance.local_temp(),
    )
    assert not result.success


def test_mem_storage_error_pipeline_multiprocess():
    result = execute_pipeline(
        ExecutionTargetHandle.for_pipeline_fn(define_diamond_pipeline).build_pipeline_definition(),
        environment_dict={'execution': {'multiprocess': {}}},
        instance=DagsterInstance.local_temp(),
        raise_on_error=False,
    )
    assert not result.success
    assert len(result.event_list) == 1
    assert result.event_list[0].is_failure


def test_invalid_instance():
    result = execute_pipeline(
        ExecutionTargetHandle.for_pipeline_fn(define_diamond_pipeline).build_pipeline_definition(),
        environment_dict={'storage': {'filesystem': {}}, 'execution': {'multiprocess': {}}},
        instance=DagsterInstance.ephemeral(),
        raise_on_error=False,
    )
    assert not result.success
    assert len(result.event_list) == 1
    assert result.event_list[0].is_failure
    assert (
        result.event_list[0].pipeline_init_failure_data.error.cls_name
        == 'DagsterUnmetExecutorRequirementsError'
    )
    assert 'non-ephemeral instance' in result.event_list[0].pipeline_init_failure_data.error.message


def test_no_handle():
    result = execute_pipeline(
        define_diamond_pipeline(),
        environment_dict={'storage': {'filesystem': {}}, 'execution': {'multiprocess': {}}},
        instance=DagsterInstance.ephemeral(),
        raise_on_error=False,
    )
    assert not result.success
    assert len(result.event_list) == 1
    assert result.event_list[0].is_failure
    assert (
        result.event_list[0].pipeline_init_failure_data.error.cls_name
        == 'DagsterUnmetExecutorRequirementsError'
    )
    assert 'ExecutionTargetHandle' in result.event_list[0].pipeline_init_failure_data.error.message


def test_solid_subset():
    pipe = ExecutionTargetHandle.for_pipeline_python_file(
        __file__, 'define_diamond_pipeline'
    ).build_pipeline_definition()

    result = execute_pipeline(pipe, preset='just_adder', instance=DagsterInstance.local_temp())

    assert result.success

    assert result.result_for_solid('adder').output_value() == 2


def define_subdag_pipeline():
    @solid(config=Field(String))
    def waiter(context):
        done = False
        while not done:
            time.sleep(0.15)
            if os.path.isfile(context.solid_config):
                return

    @solid(
        input_defs=[InputDefinition('after', Nothing)], config=Field(String),
    )
    def writer(context):
        with open(context.solid_config, 'w') as fd:
            fd.write('1')
        return

    @lambda_solid(
        input_defs=[InputDefinition('after', Nothing)], output_def=OutputDefinition(Nothing),
    )
    def noop():
        pass

    @pipeline
    def separate():
        waiter()
        a = noop.alias('noop_1')()
        b = noop.alias('noop_2')(a)
        c = noop.alias('noop_3')(b)
        writer(c)

    return separate


def test_separate_sub_dags():
    pipe = ExecutionTargetHandle.for_pipeline_python_file(
        __file__, 'define_subdag_pipeline'
    ).build_pipeline_definition()

    with safe_tempfile_path() as filename:
        result = execute_pipeline(
            pipe,
            environment_dict={
                'storage': {'filesystem': {}},
                'execution': {'multiprocess': {'config': {'max_concurrent': 2}}},
                'solids': {'waiter': {'config': filename}, 'writer': {'config': filename},},
            },
            instance=DagsterInstance.local_temp(),
        )

    assert result.success

    # this test is to ensure that the chain of noop -> noop -> noop -> writer is not blocked by waiter
    order = [str(event.solid_handle) for event in result.step_event_list if event.is_step_success]

    # the writer and waiter my finish in different orders so just ensure the proceeding chain
    assert order[0:3] == ['noop_1', 'noop_2', 'noop_3']



from dagster.cli.load_handle import _cli_load_invariant, handle_for_pipeline_cli_args
from dagster.core.definitions.container import get_active_repository_data_from_image
from dagster.core.snap import PipelineSnapshot
from dagster.seven import is_module_available


def get_pipeline_snapshot_from_cli_args(cli_args):
    _cli_load_invariant(cli_args.get('pipeline_name') is not None)

    if cli_args.get('image'):
        _cli_load_invariant(
            is_module_available('docker'),
            msg='--image is not supported without dagster[docker] or the Python package docker installed.',
        )
        active_repo_data = get_active_repository_data_from_image(cli_args.get('image'))
        return active_repo_data.get_pipeline_snapshot(cli_args.get('pipeline_name')[0])
    else:
        pipeline_definition = handle_for_pipeline_cli_args(cli_args).build_pipeline_definition()
        return PipelineSnapshot.from_pipeline_def(pipeline_definition)



from dagster import InputDefinition, OutputDefinition, solid
from dagster.core.snap.solid import build_core_solid_def_snap
from dagster.serdes import deserialize_json_to_dagster_namedtuple, serialize_dagster_namedtuple


def test_basic_solid_definition():
    @solid
    def noop_solid(_):
        pass

    solid_snap = build_core_solid_def_snap(noop_solid)

    assert solid_snap
    assert (
        deserialize_json_to_dagster_namedtuple(serialize_dagster_namedtuple(solid_snap))
        == solid_snap
    )


def test_solid_definition_kitchen_sink():
    @solid(
        input_defs=[
            InputDefinition('arg_one', str, description='desc1'),
            InputDefinition('arg_two', int),
        ],
        output_defs=[
            OutputDefinition(name='output_one', dagster_type=str),
            OutputDefinition(
                name='output_two', dagster_type=int, description='desc2', is_required=False
            ),
        ],
        config={'foo': int},
        description='a description',
        tags={'a_tag': 'yup'},
        required_resource_keys={'a_resource'},
    )
    def kitchen_sink_solid(_, arg_two, arg_one):  # out of order to test positional_inputs
        assert arg_one
        assert arg_two
        raise Exception('should not execute')

    kitchen_sink_solid_snap = build_core_solid_def_snap(kitchen_sink_solid)

    assert kitchen_sink_solid_snap
    assert kitchen_sink_solid_snap.name == 'kitchen_sink_solid'
    assert len(kitchen_sink_solid_snap.input_def_snaps) == 2
    assert [inp.name for inp in kitchen_sink_solid_snap.input_def_snaps] == ['arg_one', 'arg_two']
    assert [inp.dagster_type_key for inp in kitchen_sink_solid_snap.input_def_snaps] == [
        'String',
        'Int',
    ]

    assert kitchen_sink_solid_snap.get_input_snap('arg_one').description == 'desc1'

    assert [out.name for out in kitchen_sink_solid_snap.output_def_snaps] == [
        'output_one',
        'output_two',
    ]

    assert [out.dagster_type_key for out in kitchen_sink_solid_snap.output_def_snaps] == [
        'String',
        'Int',
    ]

    assert kitchen_sink_solid_snap.get_output_snap('output_two').description == 'desc2'
    assert kitchen_sink_solid_snap.get_output_snap('output_two').is_required is False

    assert (
        kitchen_sink_solid_snap.config_field_snap.type_key
        == kitchen_sink_solid.config_field.config_type.key
    )

    assert kitchen_sink_solid_snap.required_resource_keys == ['a_resource']
    assert kitchen_sink_solid_snap.tags == {'a_tag': 'yup'}
    assert kitchen_sink_solid.positional_inputs == ['arg_two', 'arg_one']

    assert (
        deserialize_json_to_dagster_namedtuple(
            serialize_dagster_namedtuple(kitchen_sink_solid_snap)
        )
        == kitchen_sink_solid_snap
    )



import re

import pytest

from dagster import (
    DagsterInvalidConfigDefinitionError,
    DagsterInvalidDefinitionError,
    DependencyDefinition,
    Field,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    ResourceDefinition,
    SolidDefinition,
    solid,
)
from dagster.core.utility_solids import define_stub_solid


def solid_a_b_list():
    return [
        SolidDefinition(
            name='A',
            input_defs=[],
            output_defs=[OutputDefinition()],
            compute_fn=lambda _context, _inputs: None,
        ),
        SolidDefinition(
            name='B',
            input_defs=[InputDefinition('b_input')],
            output_defs=[],
            compute_fn=lambda _context, _inputs: None,
        ),
    ]


def test_create_pipeline_with_bad_solids_list():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match='"solids" arg to pipeline "a_pipeline" is not a list. Got',
    ):
        PipelineDefinition(
            name='a_pipeline', solid_defs=define_stub_solid('stub', [{'a key': 'a value'}])
        )


def test_circular_dep():
    with pytest.raises(DagsterInvalidDefinitionError, match='circular reference'):
        PipelineDefinition(
            solid_defs=solid_a_b_list(),
            dependencies={'A': {}, 'B': {'b_input': DependencyDefinition('B')}},
        )


def test_from_solid_not_there():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='solid "NOTTHERE" in dependency dictionary not found'
    ):
        PipelineDefinition(
            solid_defs=solid_a_b_list(),
            dependencies={
                'A': {},
                'B': {'b_input': DependencyDefinition('A')},
                'NOTTHERE': {'b_input': DependencyDefinition('A')},
            },
        )


def test_from_non_existant_input():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='solid "B" does not have input "not_an_input"'
    ):
        PipelineDefinition(
            solid_defs=solid_a_b_list(),
            dependencies={'B': {'not_an_input': DependencyDefinition('A')}},
        )


def test_to_solid_not_there():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='solid "NOTTHERE" not found in solid list'
    ):
        PipelineDefinition(
            solid_defs=solid_a_b_list(),
            dependencies={'A': {}, 'B': {'b_input': DependencyDefinition('NOTTHERE')}},
        )


def test_to_solid_output_not_there():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='solid "A" does not have output "NOTTHERE"'
    ):
        PipelineDefinition(
            solid_defs=solid_a_b_list(),
            dependencies={'B': {'b_input': DependencyDefinition('A', output='NOTTHERE')}},
        )


def test_invalid_item_in_solid_list():
    with pytest.raises(
        DagsterInvalidDefinitionError, match="Invalid item in solid list: 'not_a_solid'"
    ):
        PipelineDefinition(solid_defs=['not_a_solid'])


def test_one_layer_off_dependencies():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match="Received a IDependencyDefinition one layer too high under key B",
    ):
        PipelineDefinition(
            solid_defs=solid_a_b_list(), dependencies={'B': DependencyDefinition('A')}
        )


def test_malformed_dependencies():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match='Expected IDependencyDefinition for solid "B" input "b_input"',
    ):
        PipelineDefinition(
            solid_defs=solid_a_b_list(),
            dependencies={'B': {'b_input': {'b_input': DependencyDefinition('A')}}},
        )


def test_list_dependencies():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='The expected type for "dependencies" is dict'
    ):
        PipelineDefinition(solid_defs=solid_a_b_list(), dependencies=[])


def test_pass_unrelated_type_to_field_error_solid_definition():

    with pytest.raises(DagsterInvalidConfigDefinitionError) as exc_info:

        @solid(config='nope')
        def _a_solid(_context):
            pass

    assert str(exc_info.value).startswith(
        "Error defining config. Original value passed: 'nope'. 'nope' cannot be resolved."
    )


def test_pass_unrelated_type_to_field_error_resource_definition():
    with pytest.raises(DagsterInvalidConfigDefinitionError) as exc_info:
        ResourceDefinition(resource_fn=lambda: None, config='wut')

    assert str(exc_info.value).startswith(
        "Error defining config. Original value passed: 'wut'. 'wut' cannot be resolved."
    )


def test_pass_unrelated_type_in_nested_field_error_resource_definition():
    with pytest.raises(DagsterInvalidConfigDefinitionError) as exc_info:
        ResourceDefinition(resource_fn=lambda: None, config={'field': {'nested_field': 'wut'}})
    assert str(exc_info.value).startswith('Error')

    assert str(exc_info.value).startswith(
        "Error defining config. Original value passed: {'field': {'nested_field': 'wut'}}. "
        "Error at stack path :field:nested_field. 'wut' cannot be resolved."
    )


def test_pass_incorrect_thing_to_field():
    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        Field('nope')

    assert str(exc_info.value) == (
        'Attempted to pass \'nope\' to a Field that expects a valid dagster type '
        'usable in config (e.g. Dict, Int, String et al).'
    )


def test_bad_output_definition():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match=re.escape(
            'Invalid type: dagster_type must be DagsterType, a python scalar, or '
            'a python type that has been marked usable as a dagster type via '
            '@usable_dagster_type or make_python_type_usable_as_dagster_type: '
            'got foo'
        ),
    ):
        _output = OutputDefinition('foo')

    # Test the case where the object is not hashable
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match=re.escape(
            'Invalid type: dagster_type must be DagsterType, a python scalar, or '
            'a python type that has been marked usable as a dagster type via '
            '@usable_dagster_type or make_python_type_usable_as_dagster_type: '
            'got {\'foo\': \'bar\'}, which isn\'t hashable. '
            'Did you pass an instance of a type instead of the type?'
        ),
    ):
        _output = OutputDefinition({'foo': 'bar'})

    # Test the case where the object throws in __nonzero__, e.g. pandas.DataFrame
    class Exotic(object):
        def __nonzero__(self):
            raise ValueError('Love too break the core Python APIs in widely-used libraries')

    with pytest.raises(
        DagsterInvalidDefinitionError,
        match=re.escape(
            'Invalid type: dagster_type must be DagsterType, a python scalar, or '
            'a python type that has been marked usable as a dagster type via '
            '@usable_dagster_type or make_python_type_usable_as_dagster_type: '
            'got <dagster_tests.core_tests.definitions_tests.test_definition_errors'
        )
        + '('  # py27
        + re.escape('.test_bad_output_definition.<locals>')
        + ')?'
        + re.escape('.Exotic object'),
    ):
        _output = OutputDefinition(Exotic())


def test_solid_tags():
    @solid(tags={'good': {'ok': 'fine'}})
    def _fine_tags(_):
        pass

    class X:
        pass

    with pytest.raises(
        DagsterInvalidDefinitionError, match='Could not JSON encode value',
    ):

        @solid(tags={'bad': X()})
        def _bad_tags(_):
            pass

    with pytest.raises(
        DagsterInvalidDefinitionError,
        match=r'JSON encoding "\[1, 2\]" of value "\(1, 2\)" is not equivalent to original value',
    ):

        @solid(tags={'set_comes_back_as_dict': (1, 2)})
        def _also_bad_tags(_):
            pass



import sys
from collections import namedtuple
from contextlib import contextmanager

from dagster import check
from dagster.core.definitions import ExecutionTargetHandle, PipelineDefinition
from dagster.core.definitions.resource import ScopedResourcesBuilder
from dagster.core.definitions.system_storage import SystemStorageData
from dagster.core.engine.init import InitExecutorContext
from dagster.core.errors import DagsterError
from dagster.core.events import DagsterEvent, PipelineInitFailureData
from dagster.core.execution.config import ExecutorConfig
from dagster.core.execution.memoization import validate_reexecution_memoization
from dagster.core.execution.plan.plan import ExecutionPlan
from dagster.core.execution.resources_init import (
    get_required_resource_keys_to_init,
    resource_initialization_manager,
)
from dagster.core.instance import DagsterInstance
from dagster.core.log_manager import DagsterLogManager
from dagster.core.storage.init import InitSystemStorageContext
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.core.storage.type_storage import construct_type_storage_plugin_registry
from dagster.core.system_config.objects import EnvironmentConfig
from dagster.loggers import default_loggers, default_system_loggers
from dagster.utils import EventGenerationManager, merge_dicts
from dagster.utils.error import serializable_error_info_from_exc_info

from .context.logger import InitLoggerContext
from .context.system import SystemPipelineExecutionContext, SystemPipelineExecutionContextData


def construct_system_storage_data(storage_init_context):
    return storage_init_context.system_storage_def.system_storage_creation_fn(storage_init_context)


def system_storage_def_from_config(mode_definition, environment_config):
    for system_storage_def in mode_definition.system_storage_defs:
        if system_storage_def.name == environment_config.storage.system_storage_name:
            return system_storage_def

    check.failed(
        'Could not find storage mode {}. Should have be caught by config system'.format(
            environment_config.storage.system_storage_name
        )
    )


def construct_executor_config(executor_init_context):
    return executor_init_context.executor_def.executor_creation_fn(executor_init_context)


def executor_def_from_config(mode_definition, environment_config):
    for executor_def in mode_definition.executor_defs:
        if executor_def.name == environment_config.execution.execution_engine_name:
            return executor_def

    check.failed(
        'Could not find executor {}. Should have be caught by config system'.format(
            environment_config.execution.execution_engine_name
        )
    )


# This represents all the data that is passed *into* context creation process.
# The remainder of the objects generated (e.g. loggers, resources) are created
# using user-defined code that may fail at runtime and result in the emission
# of a pipeline init failure events. The data in this object are passed all
# over the place during the context creation process so grouping here for
# ease of argument passing etc.
class ContextCreationData(
    namedtuple(
        '_ContextCreationData',
        'pipeline environment_config pipeline_run mode_def system_storage_def '
        'execution_target_handle executor_def instance resource_keys_to_init',
    )
):
    @property
    def pipeline_def(self):
        return self.pipeline.get_definition()


def create_context_creation_data(
    execution_plan, environment_dict, pipeline_run, instance,
):
    pipeline_def = execution_plan.pipeline.get_definition()
    environment_config = EnvironmentConfig.build(
        pipeline_def, environment_dict, mode=pipeline_run.mode
    )

    mode_def = pipeline_def.get_mode_definition(pipeline_run.mode)
    system_storage_def = system_storage_def_from_config(mode_def, environment_config)
    executor_def = executor_def_from_config(mode_def, environment_config)

    execution_target_handle, _ = ExecutionTargetHandle.get_handle(pipeline_def)

    return ContextCreationData(
        pipeline=execution_plan.pipeline,
        environment_config=environment_config,
        pipeline_run=pipeline_run,
        mode_def=mode_def,
        system_storage_def=system_storage_def,
        execution_target_handle=execution_target_handle,
        executor_def=executor_def,
        instance=instance,
        resource_keys_to_init=get_required_resource_keys_to_init(
            execution_plan, system_storage_def
        ),
    )


def pipeline_initialization_manager(
    execution_plan,
    environment_dict,
    pipeline_run,
    instance,
    scoped_resources_builder_cm=None,
    system_storage_data=None,
    raise_on_error=False,
):
    scoped_resources_builder_cm = check.opt_callable_param(
        scoped_resources_builder_cm,
        'scoped_resources_builder_cm',
        default=resource_initialization_manager,
    )
    generator = pipeline_initialization_event_generator(
        execution_plan,
        environment_dict,
        pipeline_run,
        instance,
        scoped_resources_builder_cm,
        system_storage_data,
        raise_on_error,
    )
    return EventGenerationManager(generator, SystemPipelineExecutionContext, raise_on_error)


def pipeline_initialization_event_generator(
    execution_plan,
    environment_dict,
    pipeline_run,
    instance,
    scoped_resources_builder_cm,
    system_storage_data=None,
    raise_on_error=False,
):
    execution_plan = check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)
    pipeline_def = execution_plan.pipeline.get_definition()

    environment_dict = check.dict_param(environment_dict, 'environment_dict', key_type=str)
    pipeline_run = check.inst_param(pipeline_run, 'pipeline_run', PipelineRun)
    instance = check.inst_param(instance, 'instance', DagsterInstance)

    scoped_resources_builder_cm = check.callable_param(
        scoped_resources_builder_cm, 'scoped_resources_builder_cm'
    )
    system_storage_data = check.opt_inst_param(
        system_storage_data, 'system_storage_data', SystemStorageData
    )
    raise_on_error = check.bool_param(raise_on_error, 'raise_on_error')

    pipeline_context = None
    resources_manager = None

    try:
        context_creation_data = create_context_creation_data(
            execution_plan, environment_dict, pipeline_run, instance,
        )
        executor_config = create_executor_config(context_creation_data)
        log_manager = create_log_manager(context_creation_data)
        resources_manager = scoped_resources_builder_cm(
            execution_plan,
            context_creation_data.environment_config,
            context_creation_data.pipeline_run,
            log_manager,
            context_creation_data.resource_keys_to_init,
        )
        for event in resources_manager.generate_setup_events():
            yield event
        scoped_resources_builder = check.inst(
            resources_manager.get_object(), ScopedResourcesBuilder
        )
        system_storage_data = create_system_storage_data(
            context_creation_data, system_storage_data, scoped_resources_builder
        )
        pipeline_context = construct_pipeline_execution_context(
            context_creation_data=context_creation_data,
            scoped_resources_builder=scoped_resources_builder,
            system_storage_data=system_storage_data,
            log_manager=log_manager,
            executor_config=executor_config,
            raise_on_error=raise_on_error,
        )

        _validate_plan_with_context(pipeline_context, execution_plan)

        yield pipeline_context
        for event in resources_manager.generate_teardown_events():
            yield event
    except DagsterError as dagster_error:
        if pipeline_context is None:
            user_facing_exc_info = (
                # pylint does not know original_exc_info exists is is_user_code_error is true
                # pylint: disable=no-member
                dagster_error.original_exc_info
                if dagster_error.is_user_code_error
                else sys.exc_info()
            )
            error_info = serializable_error_info_from_exc_info(user_facing_exc_info)

            yield DagsterEvent.pipeline_init_failure(
                pipeline_name=pipeline_def.name,
                failure_data=PipelineInitFailureData(error=error_info),
                log_manager=_create_context_free_log_manager(instance, pipeline_run, pipeline_def),
            )
            if resources_manager:
                for event in resources_manager.generate_teardown_events():
                    yield event
        else:
            # pipeline teardown failure
            raise dagster_error

        if raise_on_error:
            raise dagster_error


# perform any plan validation that is dependent on access to the pipeline context
def _validate_plan_with_context(pipeline_context, execution_plan):
    validate_reexecution_memoization(pipeline_context, execution_plan)


def create_system_storage_data(
    context_creation_data, system_storage_data, scoped_resources_builder
):
    check.inst_param(context_creation_data, 'context_creation_data', ContextCreationData)

    environment_config, pipeline_def, system_storage_def, pipeline_run = (
        context_creation_data.environment_config,
        context_creation_data.pipeline_def,
        context_creation_data.system_storage_def,
        context_creation_data.pipeline_run,
    )

    system_storage_data = (
        system_storage_data
        if system_storage_data
        else construct_system_storage_data(
            InitSystemStorageContext(
                pipeline_def=pipeline_def,
                mode_def=context_creation_data.mode_def,
                system_storage_def=system_storage_def,
                system_storage_config=environment_config.storage.system_storage_config,
                pipeline_run=pipeline_run,
                instance=context_creation_data.instance,
                environment_config=environment_config,
                type_storage_plugin_registry=construct_type_storage_plugin_registry(
                    pipeline_def, system_storage_def
                ),
                resources=scoped_resources_builder.build(
                    context_creation_data.system_storage_def.required_resource_keys,
                ),
            )
        )
    )

    return system_storage_data


def create_executor_config(context_creation_data):
    check.inst_param(context_creation_data, 'context_creation_data', ContextCreationData)

    return construct_executor_config(
        InitExecutorContext(
            pipeline=context_creation_data.pipeline,
            mode_def=context_creation_data.mode_def,
            executor_def=context_creation_data.executor_def,
            pipeline_run=context_creation_data.pipeline_run,
            environment_config=context_creation_data.environment_config,
            executor_config=context_creation_data.environment_config.execution.execution_engine_config,
            system_storage_def=context_creation_data.system_storage_def,
            instance=context_creation_data.instance,
        )
    )


def construct_pipeline_execution_context(
    context_creation_data,
    scoped_resources_builder,
    system_storage_data,
    log_manager,
    executor_config,
    raise_on_error,
):
    check.inst_param(context_creation_data, 'context_creation_data', ContextCreationData)
    scoped_resources_builder = check.inst_param(
        scoped_resources_builder if scoped_resources_builder else ScopedResourcesBuilder(),
        'scoped_resources_builder',
        ScopedResourcesBuilder,
    )
    check.inst_param(system_storage_data, 'system_storage_data', SystemStorageData)
    check.inst_param(log_manager, 'log_manager', DagsterLogManager)
    check.inst_param(executor_config, 'executor_config', ExecutorConfig)

    return SystemPipelineExecutionContext(
        SystemPipelineExecutionContextData(
            pipeline=context_creation_data.pipeline,
            mode_def=context_creation_data.mode_def,
            system_storage_def=context_creation_data.system_storage_def,
            pipeline_run=context_creation_data.pipeline_run,
            scoped_resources_builder=scoped_resources_builder,
            environment_config=context_creation_data.environment_config,
            instance=context_creation_data.instance,
            intermediates_manager=system_storage_data.intermediates_manager,
            file_manager=system_storage_data.file_manager,
            execution_target_handle=context_creation_data.execution_target_handle,
            executor_config=executor_config,
            raise_on_error=raise_on_error,
        ),
        log_manager=log_manager,
    )


@contextmanager
def scoped_pipeline_context(
    execution_plan,
    environment_dict,
    pipeline_run,
    instance,
    scoped_resources_builder_cm=resource_initialization_manager,
    system_storage_data=None,
    raise_on_error=False,
):
    ''' Utility context manager which acts as a very thin wrapper around
    `pipeline_initialization_manager`, iterating through all the setup/teardown events and
    discarding them.  It yields the resulting `pipeline_context`.

    Should only be used where we need to reconstruct the pipeline context, ignoring any yielded
    events (e.g. PipelineExecutionResult, dagstermill, unit tests, etc)
    '''
    check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)
    check.dict_param(environment_dict, 'environment_dict', key_type=str)
    check.inst_param(pipeline_run, 'pipeline_run', PipelineRun)
    check.inst_param(instance, 'instance', DagsterInstance)
    check.callable_param(scoped_resources_builder_cm, 'scoped_resources_builder_cm')
    check.opt_inst_param(system_storage_data, 'system_storage_data', SystemStorageData)

    initialization_manager = pipeline_initialization_manager(
        execution_plan,
        environment_dict,
        pipeline_run,
        instance,
        scoped_resources_builder_cm=scoped_resources_builder_cm,
        system_storage_data=system_storage_data,
        raise_on_error=raise_on_error,
    )
    for _ in initialization_manager.generate_setup_events():
        pass

    try:
        yield check.inst(initialization_manager.get_object(), SystemPipelineExecutionContext)
    finally:
        for _ in initialization_manager.generate_teardown_events():
            pass


def create_log_manager(context_creation_data):
    check.inst_param(context_creation_data, 'context_creation_data', ContextCreationData)

    pipeline_def, mode_def, environment_config, pipeline_run = (
        context_creation_data.pipeline_def,
        context_creation_data.mode_def,
        context_creation_data.environment_config,
        context_creation_data.pipeline_run,
    )

    loggers = []
    for logger_key, logger_def in mode_def.loggers.items() or default_loggers().items():
        if logger_key in environment_config.loggers:
            loggers.append(
                logger_def.logger_fn(
                    InitLoggerContext(
                        environment_config.loggers.get(logger_key, {}).get('config'),
                        pipeline_def,
                        logger_def,
                        pipeline_run.run_id,
                    )
                )
            )

    if not loggers:
        for (logger_def, logger_config) in default_system_loggers():
            loggers.append(
                logger_def.logger_fn(
                    InitLoggerContext(logger_config, pipeline_def, logger_def, pipeline_run.run_id)
                )
            )

    # should this be first in loggers list?
    loggers.append(context_creation_data.instance.get_logger())

    return DagsterLogManager(
        run_id=pipeline_run.run_id,
        logging_tags=get_logging_tags(pipeline_run, context_creation_data.pipeline_def),
        loggers=loggers,
    )


def _create_context_free_log_manager(instance, pipeline_run, pipeline_def):
    '''In the event of pipeline initialization failure, we want to be able to log the failure
    without a dependency on the ExecutionContext to initialize DagsterLogManager.
    Args:
        pipeline_run (dagster.core.storage.pipeline_run.PipelineRun)
        pipeline_def (dagster.definitions.PipelineDefinition)
    '''
    check.inst_param(instance, 'instance', DagsterInstance)
    check.inst_param(pipeline_run, 'pipeline_run', PipelineRun)
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)

    loggers = [instance.get_logger()]
    # Use the default logger
    for (logger_def, logger_config) in default_system_loggers():
        loggers += [
            logger_def.logger_fn(
                InitLoggerContext(logger_config, pipeline_def, logger_def, pipeline_run.run_id)
            )
        ]

    return DagsterLogManager(
        pipeline_run.run_id, get_logging_tags(pipeline_run, pipeline_def), loggers
    )


def get_logging_tags(pipeline_run, pipeline_def):
    check.opt_inst_param(pipeline_run, 'pipeline_run', PipelineRun)
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)

    return merge_dicts(
        {'pipeline': pipeline_def.name},
        pipeline_run.tags if pipeline_run and pipeline_run.tags else {},
    )



import json
import os

import pytest
import yaml
from click.testing import CliRunner

from dagster import seven
from dagster.cli.pipeline import pipeline_execute_command
from dagster.core.instance import DagsterInstance
from dagster.core.telemetry import get_dir_from_dagster_home
from dagster.core.test_utils import environ
from dagster.utils import pushd, script_relative_path

EXPECTED_KEYS = set(
    ['action', 'client_time', 'elapsed_time', 'event_id', 'instance_id', 'metadata',]
)


def path_to_tutorial_file(path):
    return script_relative_path(
        os.path.join('../../../../examples/dagster_examples/intro_tutorial/', path)
    )


@pytest.mark.skipif(
    os.name == 'nt', reason="TemporaryDirectory disabled for win because of event.log contention"
)
def test_dagster_telemetry_enabled(caplog):
    with seven.TemporaryDirectory() as temp_dir:
        with environ({'DAGSTER_HOME': temp_dir}):
            with open(os.path.join(temp_dir, 'dagster.yaml'), 'w') as fd:
                yaml.dump({'telemetry': {'enabled': True}}, fd, default_flow_style=False)

            DagsterInstance.local_temp(temp_dir)
            runner = CliRunner(env={'DAGSTER_HOME': temp_dir})
            with pushd(path_to_tutorial_file('')):
                result = runner.invoke(
                    pipeline_execute_command,
                    [
                        '-f',
                        path_to_tutorial_file('hello_cereal.py'),
                        '-n',
                        'hello_cereal_pipeline',
                    ],
                )

                for record in caplog.records:
                    message = json.loads(record.getMessage())
                    assert set(message.keys()) == EXPECTED_KEYS

                assert len(caplog.records) == 4
                assert result.exit_code == 0


@pytest.mark.skipif(
    os.name == 'nt', reason="TemporaryDirectory disabled for win because of event.log contention"
)
def test_dagster_telemetry_disabled(caplog):
    with seven.TemporaryDirectory() as temp_dir:
        with environ({'DAGSTER_HOME': temp_dir}):
            with open(os.path.join(temp_dir, 'dagster.yaml'), 'w') as fd:
                yaml.dump({'telemetry': {'enabled': False}}, fd, default_flow_style=False)

            DagsterInstance.local_temp(temp_dir)

            runner = CliRunner(env={'DAGSTER_HOME': temp_dir})
            with pushd(path_to_tutorial_file('')):
                result = runner.invoke(
                    pipeline_execute_command,
                    [
                        '-f',
                        path_to_tutorial_file('hello_cereal.py'),
                        '-n',
                        'hello_cereal_pipeline',
                    ],
                )

            assert not os.path.exists(os.path.join(get_dir_from_dagster_home('logs'), 'event.log'))
            assert len(caplog.records) == 0
            assert result.exit_code == 0


@pytest.mark.skipif(
    os.name == 'nt', reason="TemporaryDirectory disabled for win because of event.log contention"
)
def test_dagster_telemetry_unset(caplog):
    with seven.TemporaryDirectory() as temp_dir:
        with environ({'DAGSTER_HOME': temp_dir}):
            with open(os.path.join(temp_dir, 'dagster.yaml'), 'w') as fd:
                yaml.dump({}, fd, default_flow_style=False)

            DagsterInstance.local_temp(temp_dir)
            runner = CliRunner(env={'DAGSTER_HOME': temp_dir})
            with pushd(path_to_tutorial_file('')):
                result = runner.invoke(
                    pipeline_execute_command,
                    [
                        '-f',
                        path_to_tutorial_file('hello_cereal.py'),
                        '-n',
                        'hello_cereal_pipeline',
                    ],
                )

                for record in caplog.records:
                    message = json.loads(record.getMessage())
                    assert set(message.keys()) == EXPECTED_KEYS

                assert len(caplog.records) == 4
                assert result.exit_code == 0
